/*
	convert files from 3d studio binary (.3ds) to ascii inventor 2.1 format

	Copyright (c) deleet 2003, Johannes Zander
	http://deleet.de/io/
	
	3ds2iv is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	3ds2iv is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with 3ds2iv; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <math.h>
#include <float.h>
#include <vector>
#include <string>

#ifdef __GNUC__
	#if( __GNUC__ < 3 || ( __GNUC__ == 3 && __GNUC_MINOR__ < 1 )) // gcc version < 3.1.0 ?
		#include <hash_map>
		#define STL_EXT  std
	#else
		#include <ext/hash_map>
		#define STL_EXT  __gnu_cxx
	#endif // GCC_VERSION
#else
	#include <hash_map>
	#define STL_EXT  std
#endif // __GNUC__


#define gTINY                (1e-6f)

#define _gStringize( L )     #L
#define _gMakeString( M, L ) M( L )
#define _gLineString         _gMakeString( _gStringize, __LINE__ )
#define _gError( str )       { printf( __FILE__ "(" _gLineString "): " str ); exit(1); }
#define _gCheck( x, str )    if( !( x )) _gError( str )
#define _gVerify( x, str )   if( !( x )) _gError( str )

#ifdef _MSC_VER // MS's for-scope is plain wrong - fix it!
	#define for if( false ){}else for
#endif

template <class T>  inline void _swap( T &a, T &b ) { const T temp = a; a=b; b=temp; }


//////
// unified signed/unsigned 8/16/32/64 bit types

#if( defined( WIN32 ) || defined( _WIN32 ))
	typedef unsigned __int64 u64;
	typedef unsigned __int32 u32;
	typedef unsigned __int16 u16;
	typedef unsigned __int8  u8;
	
	typedef signed   __int64 s64;
	typedef signed   __int32 s32;
	typedef signed   __int16 s16;
	typedef signed   __int8  s8;
#elif( defined( __GNUC__ ) || defined( __MACOSX__ ))
	#include <inttypes.h>
	typedef uint64_t u64;
	typedef uint32_t u32;
	typedef uint16_t u16;
	typedef uint8_t  u8;
	
	typedef int64_t  s64;
	typedef int32_t  s32;
	typedef int16_t  s16;
	typedef int8_t   s8;
#elif
	#error "uh - no idea what the int types could be!"
#endif


//////
// 3D Vector class

	struct Vec3 {
		float x,y,z;
		
		inline Vec3() {}
		inline Vec3( float _x, float _y, float _z ) : x(_x),   y(_y),   z(_z)   {}
		inline Vec3( const float *v )               : x(v[0]), y(v[1]), z(v[2]) {}
		inline Vec3( const Vec3 &v )                : x(v.x),  y(v.y),  z(v.z)  {}
	
		inline Vec3& operator()( float _x, float _y, float _z ) { x=_x;   y=_y;    z=_z;   return( *this ); }
		inline Vec3& operator()( const float *v )               { x=v[0]; y=v[1];  z=v[2]; return( *this ); }
		inline Vec3& operator()( const Vec3 &v )                { x=v.x;  y=v.y;   z=v.z;  return( *this ); }
		
		inline Vec3& operator=( const Vec3& v ) {
			return( operator()( v ));
		}
	
		inline operator float*() {
			return( &x );
		}
		
		inline operator const float*() const {
			return( &x );
		}
		
		inline float& operator []( u32 i ) {
			return( *( &x + i ));
		}
	
		inline float operator []( u32 i ) const {
			return( *( &x + i ));
		}
	
		bool operator==( const Vec3 &v ) const {
			return( fabs( x-v.x ) <= gTINY && fabs( y-v.y ) <= gTINY && fabs( z-v.z ) <= gTINY );
		}
	
		inline bool operator!=( const Vec3 &v ) const {
			return( !((*this)==v ));
		}
	
		Vec3& operator+=( const Vec3 &v ) {
			x += v.x;
			y += v.y;
			z += v.z;
			return( *this );
		}
	
		Vec3& operator-=( const Vec3 &v ) {
			x -= v.x;
			y -= v.y;
			z -= v.z;
			return( *this );
		}
	
		Vec3& operator*=( float s ) {
			x *= s;
			y *= s;
			z *= s;
			return( *this );
		}
		
		inline Vec3& operator/=( float s ) {
			return( *this *= (1.f / s ));
		}
	
		Vec3 operator-() const {
			return( Vec3( -x, -y, -z ));
		}
	
		friend Vec3 operator*( const Vec3& v, float s ) {
			return( Vec3( v.x*s, v.y*s, v.z*s ));
		}
	
		friend inline Vec3 operator*( float s, const Vec3& v ) {
			return( v*s );
		}
	
		friend inline Vec3 operator/( const Vec3& v, float s ) {
			return( v*( 1.f/s ));
		}
		
		friend Vec3 operator+( const Vec3& v0, const Vec3& v1 ) {
			return( Vec3( v0.x+v1.x, v0.y+v1.y, v0.z+v1.z ));
		}
	
		friend Vec3 operator-( const Vec3& v0, const Vec3& v1 ) {
			return( Vec3( v0.x-v1.x, v0.y-v1.y, v0.z-v1.z ));
		}
		
		Vec3& negate() {
			x = -x;
			y = -y;
			z = -z;
			return( *this );
		}
		
		float dot( const Vec3 &v ) const { // dot product
			return( x*v.x + y*v.y + z*v.z );
		}
	
		inline float length() const {
			return( (float)sqrt( dot( *this )));
		}
	
		Vec3& normalize() { // unit length
			const float len2 = dot( *this );
			
			if( len2>0 && fabs( len2-1 )>gTINY ) (*this) /= (float)sqrt( len2 );
			return( *this );
		}
	
		Vec3& normalize( float& len ) { // unit length
			const float len2 = dot( *this );
			
			if( len2>0 ) {
				if( fabs( len2-1 )>gTINY ) (*this) /= len = (float)sqrt( len2 ); else len = 1;
			} else len = 1;
			return( *this );
		}
	
		inline Vec3 unit() const {
			return( Vec3( *this ).normalize());
		}
	
		static float distance( const Vec3 &a, const Vec3 &b ) { // length AB
			return(( a-b ).length());
		}
	
		Vec3 cross( const Vec3 &v ) const { // cross product
			return( Vec3(
				y*v.z - z*v.y,
				z*v.x - x*v.z,
				x*v.y - y*v.x
			));
		}
	
		static Vec3 cross( const Vec3 &u, const Vec3 &v ) { // cross product
			return( Vec3(
				u.y*v.z - u.z*v.y,
				u.z*v.x - u.x*v.z,
				u.x*v.y - u.y*v.x
			));
		}
	
		static Vec3& cross( const Vec3 &u, const Vec3 &v, Vec3 &out ) { // cross product
			out.x = u.y*v.z - u.z*v.y;
			out.y = u.z*v.x - u.x*v.z;
			out.z = u.x*v.y - u.y*v.x;
			return( out );
		}
	};


//////
// halfedge model

	struct Vertex;
	struct Triangle;
	struct TriangleRef;
	class  HeModel;
	
	typedef std::pair<Vertex*, Vertex*> Edge;
	
	struct HalfEdge {
		Vertex*      vert; // target vertex
		HalfEdge*    next; // next edge CCW around vertex
		HalfEdge*    loop; // next edge CCW in triangle
		HalfEdge*    pair; // other side of edge / neighbour
		TriangleRef* tri;  // triangle
	};
	
	struct Vertex {
		int       id;
		Vec3      pos, normal;
		HalfEdge* edges; // all the edges around this vertex one
	
		int valence() const {
			int val = 0;
	
			if( edges ) {
				HalfEdge* e = edges;
				do { ++val; e = e->next; } while( e && e!=edges );
			}
			return( val );
		}
	};
	
	struct Triangle {
		int idx0, idx1, idx2;
	
		Triangle( int _idx0, int _idx1, int _idx2 )
			: idx0(_idx0), idx1(_idx1), idx2(_idx2) {}
	};
	
	struct TriangleRef {
		int       id;
		Vertex   *vtx0, *vtx1, *vtx2;
		HalfEdge *loop; // edge v0-v1
		Vec3      normal;
	};
	
	class HeModel {
	private:
		struct EdgeHash {
			inline size_t operator()( const Edge& e ) const {
				return((size_t)e.first^(((size_t)e.second << 16 )&0xffff )^(((size_t)e.second >> 16 )&0xffff ));
			}
		};
	
		struct EdgeEq {
			inline size_t operator()( const Edge& a, const Edge& b ) const {
				return( a.first==b.first && a.second==b.second );
			}
		};
	
		typedef STL_EXT::hash_map<Edge, HalfEdge*, EdgeHash, EdgeEq>  EdgeMap;
		typedef std::vector<Vertex*>                                  VertexMap;
		typedef std::vector<TriangleRef*>                             TriangleMap;
	
	
		TriangleMap triangleMap;
		VertexMap   vertexMap;
		EdgeMap     edgeMap;
	
	
		static HalfEdge* newHalfEdge( const Edge& edge, TriangleRef* _tri=0 ) {
			HalfEdge* he = new HalfEdge();
	
			he->vert = edge.second;
			he->next = 0;
			he->loop = 0;
			he->pair = 0;
			he->tri  = _tri;
	
			return( he );
		}
	
	
		static Vertex* newVertex( const Vec3& _pos, int _id ) {
			Vertex* vtx = new Vertex();
	
			vtx->id    = _id;
			vtx->pos   = _pos;
			vtx->normal(0,0,0);
			vtx->edges = 0;
	
			return( vtx );
		}
		
		static TriangleRef* newTriangleRef( Vertex* _vtx0, Vertex* _vtx1, Vertex* _vtx2, int _id ) {
			_gCheck( _vtx0 && _vtx1 && _vtx2, "HeModel::newTriangleRef: NULL argument" );
			TriangleRef* tri = new TriangleRef();
	
			tri->id   = _id;
			tri->vtx0 = _vtx0;
			tri->vtx1 = _vtx1;
			tri->vtx2 = _vtx2;
			tri->loop = 0;
	
			tri->normal = Vec3::cross( _vtx1->pos - _vtx0->pos, _vtx2->pos - _vtx0->pos ).normalize();
			_vtx0->normal += tri->normal; _vtx1->normal += tri->normal; _vtx2->normal += tri->normal;
	
			return( tri );
		}
	
	
	public:
		HeModel()
			: finalized( false )
			, maxVertexId(-1), maxTriangleId(-1) {}
	
		~HeModel() {
			clear();
		}
	
		void clear() {
			for( EdgeMap::const_iterator i=edgeMap.begin(); i!=edgeMap.end(); ++i ) delete (*i).second;
			for( int i=vertexMap.size()-1;   i>=0; --i ) delete vertexMap[i];
			for( int i=triangleMap.size()-1; i>=0; --i ) delete triangleMap[i];
			edgeMap.clear();
			vertexMap.clear();
			triangleMap.clear();
			finalized = false;
			maxVertexId = maxTriangleId = -1;
		}
	
		inline bool empty() const {
			return( maxTriangleId < 0 );
		}
	
		
		inline const TriangleRef* triangle( u32 id ) const {
			return( triangleMap[id] );
		}
	
		inline const HalfEdge* edge( int from, int to ) const {
			return( findEdge( Edge( vertexMap[from], vertexMap[to] )));
		}
	
		inline const HalfEdge* edge( const Vertex* from, const Vertex* to ) const {
			return( findEdge( Edge( const_cast<Vertex*>( from ), const_cast<Vertex*>( to ))));
		}
	
		inline const Vertex* vertex( u32 id ) const {
			return( vertexMap[id] );
		}
	
		
		inline int nextVertexId() const {
			return( maxVertexId+1 );
		}
	
		inline int nextTriangleId() const {
			return( maxTriangleId+1 );
		}
	
		
		void add( const Vec3& pos, int id ) {
			_gCheck( !finalized, "HeModel::add: model already finalized" );
			_gCheck( id==nextVertexId(), "HeModel::add: vertex id not sequential" );
	
			vertexMap.push_back( newVertex( pos, id ));
			
			maxVertexId = id;
		}
	
		void add( const Triangle& tri, int id ) {
			_gCheck( !finalized, "HeModel::add: model already finalized" );
			_gCheck( vertexDefined( tri.idx0 ) && vertexDefined( tri.idx1 ) && vertexDefined( tri.idx2 ), "HeModel::add: vertex undefined" );
			_gCheck( id==nextTriangleId(), "HeModel::add: triangle id not sequential" );
			
			if( tri.idx0==tri.idx1 || tri.idx0==tri.idx2 || tri.idx1==tri.idx2 ) {
				puts("HeModel::add: ignored degenerated triangle");
				return;
			}
			TriangleRef *triRef = newTriangleRef( vertexMap[tri.idx0], vertexMap[tri.idx1], vertexMap[tri.idx2], id );
			triangleMap.push_back( triRef );
			triRef->loop  = insertEdge( Edge( triRef->vtx0, triRef->vtx1 ), triRef );
				            insertEdge( Edge( triRef->vtx1, triRef->vtx2 ), triRef );
				            insertEdge( Edge( triRef->vtx2, triRef->vtx0 ), triRef );
			maxTriangleId = id;
		}
		
		void finalize() {
			_gCheck( !finalized, "HeModel::finalize: model already finalized" );
	
			// connect halfedges
			for( int i=triangleMap.size()-1; i>=0; --i ) {
				TriangleRef* triRef = triangleMap[i];
	
				HalfEdge*    hEdge[] = { // edges CCW
					findEdge( Edge( triRef->vtx0, triRef->vtx1 )),
					findEdge( Edge( triRef->vtx1, triRef->vtx2 )),
					findEdge( Edge( triRef->vtx2, triRef->vtx0 ))
				};
				HalfEdge*    hEdgePair[] = { // matching backward edges
					findEdge( Edge( triRef->vtx1, triRef->vtx0 )),
					findEdge( Edge( triRef->vtx2, triRef->vtx1 )),
					findEdge( Edge( triRef->vtx0, triRef->vtx2 ))
				};
	
				_gCheck( hEdge[0] && hEdge[1] && hEdge[2], "HeModel::finalize: missing edge" );
//				_gCheck( hEdgePair[0] && hEdgePair[1] && hEdgePair[2], "HeModel::objekt nicht geschlossen" );
				for( int j=0; j<3; ++j ) {
					const int next = ( j+1 )%3, prev = ( j+2 )%3;
					if( !hEdgePair[j] )       // edge on border -> edge iterator not arbitrary -> change to border edge
						hEdge[prev]->vert->edges = hEdge[j];
					if( !hEdgePair[prev] ) {  // close this side
						Edge edge( hEdge[prev]->vert, hEdge[next]->vert );
						hEdge[prev]->pair    = hEdge[j]->next = insertEdge( edge );
						hEdge[j]->next->pair = hEdge[prev];
					} else
						hEdge[j]->next = hEdgePair[prev];
					if( !hEdge[j]->pair )	  // connect to pair
						hEdge[j]->pair = hEdgePair[j];
					hEdge[j]->loop = hEdge[next];
				}
			}
	
			// normalize summed normal vectors
			for( int i=vertexMap.size()-1; i>=0; --i )
				vertexMap[i]->normal.normalize();
			
			finalized = true;
		}
	
		bool writeIV( const char* filename, const char* generator="HeModel" ) const {
			if( empty()) return( true );
	
			FILE *file = fopen( filename, "wb" );
			if( file ) {
				bool                error = false;
				std::vector< int >  strip;
				computeStripOrder( strip );
	
				const int  lastVertex = vertexMap.size()-1;
				const int  lastStrip  = strip.size()-1;
	
				error |= ( 0>fprintf( file,
					"#Inventor V2.1 ascii\n#%s\n\nSeparator {\n"
					"\tShapeHints {\n\t\tvertexOrdering\tCOUNTERCLOCKWISE\n\t\tshapeType\tSOLID\n\t\tfaceType\tCONVEX\n\t}\n"
					"\tMaterial {\n\t\tdiffuseColor\t0.3451 0.3451 0.8824\n\t}\n"
					"\tCoordinate3 {\n\tpoint [\n", generator ));
					for( int i=0; i<=lastVertex; ++i )
						error |= ( 0>fprintf( file, "\t\t%f %f %f%c\n", vertex(i)->pos.x, vertex(i)->pos.y, vertex(i)->pos.z, i<lastVertex ? ',' : ']' ));
				error |= ( 0>fprintf( file,
					"\t}\n"
					"\tMaterialBinding {\n\t\tvalue	OVERALL\n\t}\n"
					"\tIndexedTriangleStripSet {\n\tcoordIndex [\n"));
					int i;
					for( i=0; i+8<=lastStrip; i+=8 )
						error |= ( 0>fprintf( file, "\t\t%i, %i, %i, %i, %i, %i, %i, %i%c\n",
							strip[i+0], strip[i+1], strip[i+2], strip[i+3],
							strip[i+4], strip[i+5], strip[i+6], strip[i+7], i+8<lastStrip ? ',' : ']' ));
					if( i!=lastStrip )
						for( error |= ( 0>fprintf( file, "\t\t" )); i<=lastStrip; ++i )
							error |= ( 0>fprintf( file, "%i%s", strip[i], i<lastStrip ? ", " : " ]\n" ));
				error |= ( 0>fprintf( file,
					"\t}\n}"));
				fclose( file );
				return( !error );
			}
			return( false );
		}
	
	
	protected:
		enum StripFlags { SF_GREEDY_PASSES=1, SF_GROW_BACKWARDS=2, SF_PRESERVE_WINDING=4 };
	
	
		bool  finalized;
		int   maxVertexId, maxTriangleId;
	
	
		inline bool vertexDefined( int id ) const {
			return( id>=0 && id<=maxVertexId );
		}
	
		inline bool triangleDefined( int id ) const {
			return( id>=0 && id<=maxTriangleId );
		}
	
		inline HalfEdge* findEdge( const Edge& e ) const {
			EdgeMap::const_iterator i = edgeMap.find( e );
	
			return( i!=edgeMap.end() ? (*i).second : 0 );
		}
	
		HalfEdge* insertEdge( const Edge& e, TriangleRef* tri=0 ) {
			_gCheck( !findEdge( e ), "HeModel::insertEdge: not a 2-manifold or inconsistent face orientation" );
	
			HalfEdge* hEdge = newHalfEdge( e, tri );
			edgeMap[e] = hEdge;
			if( !e.first->edges ) e.first->edges = hEdge;
			return( hEdge );
		}
	
		// striping code inspired by source from pierre terdiman
	
		static inline int oppositeVertexId( const HalfEdge* edge ) {
			return( edge->loop->vert->id );
		}
	
		static inline const HalfEdge* nextEdgeInStrip( const HalfEdge* edge, int lastVertex ) {
			if( edge->vert->id == lastVertex ) return( edge->loop->pair );
			edge = edge->loop->loop;
			if( edge->vert->id == lastVertex ) return( edge->pair );
			return( 0 );
		}
	
		static void trackStrip( const HalfEdge* edge, std::vector< int >  &strip, std::vector< bool >  &visited, bool outputFirst=true ) {
			if( !edge->tri || visited[edge->tri->id]) return;
			int lastVertex = edge->vert->id;
			if( outputFirst ) {
				strip.push_back( edge->pair->vert->id );
				strip.push_back( lastVertex );
			}
	
			do {
				int nextVertex = oppositeVertexId( edge );
	
				strip.push_back( nextVertex );
				visited[edge->tri->id] = true;
				edge                   = nextEdgeInStrip( edge, lastVertex );
				lastVertex             = nextVertex;
			} while( edge && edge->tri && !visited[edge->tri->id]);
		}
		
		void computeStripOrder( std::vector< int > &finalStrip, int flags=SF_GROW_BACKWARDS|SF_PRESERVE_WINDING ) const {
			const int            triangleNum = triangleMap.size();
			std::vector< bool >  visited( triangleNum, false );
			std::vector< bool >  visitedCopy[3];
			std::vector< int >   strip[3];
			int                  firstLength[3];
			const int            passes = flags&SF_GREEDY_PASSES ? 3 : 1;
	
			finalStrip.clear(); finalStrip.reserve( triangleNum+1 );
	
			for( int i=0; true; ++i ) {
				while( visited[i] && i<triangleNum ) ++i;
				if( i==triangleNum) break;
	
				// compute best strip from all three possibilities
				HalfEdge* ref[3] = { triangleMap[i]->loop }; ref[1] = ref[0]->loop; ref[2] = ref[1]->loop;
				for( int j=0; j<passes; ++j ) {
					if( flags&SF_GREEDY_PASSES ) visitedCopy[j] = visited; else visitedCopy[j].swap( visited );
					strip[j].clear(); strip[j].reserve( triangleNum+2+1+2 );
	
					// track 1st part of strip
					trackStrip( ref[j], strip[j], visitedCopy[j], true );
					firstLength[j] = strip[j].size();
	
					if( flags&SF_GROW_BACKWARDS ) {
						// reverse it
						std::reverse( strip[j].begin(), strip[j].end());
	
						// and track 2nd part of strip
						trackStrip( ref[j]->pair, strip[j], visitedCopy[j], false );
					}
				}
	
				// get the longest strip
				int best = 0;
				for( int j=1; j<passes; ++j ) if( strip[j].size() > strip[best].size()) best = j;
	
				visited.swap( visitedCopy[best]);
	
				// if needed flip strip to preserve face orientation
				if(( flags&SF_PRESERVE_WINDING ) && firstLength[best]&1 ) {
					if( strip[best].size()==3 || strip[best].size()==4 ) // triangles and quads are easy...
						_swap( strip[best][1], strip[best][2]);
					else {
						std::reverse( strip[best].begin(), strip[best].end());
						int secondLength = strip[best].size() - firstLength[best];
						if( secondLength&1 ) strip[best].insert( strip[best].begin(), strip[best][0]); // replicate first
					}
				}
	
				// append to master strip
				finalStrip.insert( finalStrip.end(), strip[best].begin(), strip[best].end());
				finalStrip.push_back(-1);
			}
		}
	
		// disallow
		HeModel( const HeModel& );
		HeModel& operator = ( const HeModel& );
	};


//////
// 3ds loader inspired by terry caton's source

	class Load3ds {
	public:
		Load3ds( const char* filename, HeModel& out )
			: file( ::fopen( filename, "rb" ))
			, model( out ) {
			_gCheck( file, "Load3ds::cannot open file" );
            
			Block primary;
			_gVerify( readBlockHeader( primary ) && primary.type==BT_PRIMARY, "Load3ds::not a 3ds file" );
			_gVerify( readBlock( primary ), "Load3ds::error in 3ds file" );
			model.finalize();
		}

		~Load3ds() {
			if( file ) ::fclose( file );
		}

	protected:
		enum BlockType {
			BT_PRIMARY    = 0x4D4D,
			BT_VERSION    = 0x0002, 
			BT_EDIT3DS    = 0x3D3D, BT_MESHVERSION = 0x3D3E, 
			BT_EDITOBJECT = 0x4000, BT_OBJTRIMESH  = 0x4100, BT_TRIVERT = 0x4110, BT_TRIFACE = 0x4120
		};

		struct Block {
			u16 type;
			u32 length;
			u32 read;

			Block() : read(0) {}

			inline int  left() const     { return( length-read ); }
			inline bool finished() const { return( left()==0 ); }
		};


		FILE*     file;
		HeModel&  model;

		static const u32  endianTest;
		static const u8   littleEndian;
		

		inline bool read( Block& block, u8& x ) {
			if( ::fread( &x, 1, 1, file )!=1 ) return( false );
			block.read += 1;
			return( true );
		}

		bool read( Block& block, u16& x ) {
			if( littleEndian ) {
				if( ::fread( &x, 1, 2, file )!=2 ) return( false );
			} else {
				u8 tmp[2];
				if( ::fread( tmp, 1, 2, file )!=2 ) return( false );
				x = tmp[0] + tmp[1]*0x100; // reverse
			}
			block.read += 2;
			return( true );
		}
		
		bool read( Block& block, u32& x ) {
			if( littleEndian ) {
				if( ::fread( &x, 1, 4, file )!=4 ) return( false );
			} else {
				u8 tmp[4];
				if( ::fread( tmp, 1, 4, file )!=4 ) return( false );
				x = tmp[0] + tmp[1]*0x100 + tmp[2]*0x10000 + tmp[3]*0x1000000; // reverse
			}
			block.read += 4;
			return( true );
		}

		bool read( Block& block, float& x ) {
			if( littleEndian ) {
				if( ::fread( &x, 1, 4, file )!=4 ) return( false );
			} else {
				union { u32 l; u8 i[4]; float f; } tmp;
				if( ::fread( tmp.i, 1, 4, file )!=4 ) return( false );
				tmp.l = tmp.i[0] + tmp.i[1]*0x100 + tmp.i[2]*0x10000 + tmp.i[3]*0x1000000; // reverse
				x     = tmp.f;
			}
			block.read += 4;
			return( true );
		}

		bool read( Block& block, std::string& str ) {
			if( !str.empty()) str.erase( str.begin(), str.end());
			while( !block.finished()) {
				u8 chr;
				if( !read( block, chr )) return( false );
				if( !chr ) break;
				str += (char)chr;
			}
			return( true );
		}

		inline void skipRest( Block& block ) {
			::fseek( file, block.left(), SEEK_CUR );
			block.read = block.length;
		}

		bool readBlockHeader( Block& block ) {
			block.read = 0;
			return( read( block, block.type ) && read( block, block.length ));
		}

		bool readBlock( Block& parent ) {
			while( parent.read < parent.length ) {
				Block current;

				if( !readBlockHeader( current )) return( false );
				switch( current.type ) {
				case BT_MESHVERSION:
				case BT_VERSION: {
					u16 version;
					if( !read( current, version )) return( false );
					_gCheck( version <= 0x03, "Load3ds::version > 3" );
					skipRest( current );
					} break;
				
				case BT_EDIT3DS:
					if( !readBlock( current )) return( false );
					break;

				case BT_EDITOBJECT:	{
					std::string name;
					if( !read( current, name )) return( false );
					if( !readObject( current )) return( false );
					} break;
				
				default: // unknown/not interesting
					skipRest( current );
				}
			
				parent.read += current.read;
			}
			return( true );
		}

		bool readObject( Block& parent ) {
			const int vertexIdx = model.nextVertexId();
			const int faceIdx   = model.nextTriangleId();

			while( parent.read < parent.length ) {
				Block current;

				if( !readBlockHeader( current )) return( false );
				switch( current.type ) {
				case BT_OBJTRIMESH: if( !readObject( current ))                    return( false ); break;
				case BT_TRIVERT:    if( !readVertices( current, vertexIdx ))       return( false ); break;
				case BT_TRIFACE:	if( !readFaces( current, faceIdx, vertexIdx )) return( false ); break;
				default: 			skipRest( current );
				}
			
				parent.read += current.read;
			}
			return( true );
		}

		bool readFaces( Block& block, int faceIdx, int vtxIdx ) {
			u16 num;
			u16 index[4];

			if( !read( block, num )) return( false );
			_gCheck( num*2L*4<=block.left(), "Load3ds::wrong number of triangles" );
			for( int i=0; i<num; ++i ) {
				for( int j=0; j<4; ++j ) if( !read( block, index[j])) return( false );
				model.add( Triangle( index[0]+vtxIdx, index[1]+vtxIdx, index[2]+vtxIdx ), i+faceIdx );
			}
			skipRest( block );
			return( true );
		}

		bool readVertices( Block& block, int vtxIdx ) {
			u16 num;
			float pos[3];

			if( !read( block, num )) return( false );
			_gCheck( num*4L*3<=block.left(), "Load3ds::wrong number of vertices" );
			for( int i=0; i<num; ++i ) {
				for( int j=0; j<3; ++j ) if( !read( block, pos[j])) return( false );
				model.add( Vec3( pos ), i+vtxIdx );
			}
			skipRest( block );

			return( true );
		}
	};

	const u32 Load3ds::endianTest   = 1;
	const u8  Load3ds::littleEndian = *((u8*)&Load3ds::endianTest); // detect byte ordering
	

int main( int argc, char** argv ) {
	if( argc==3 ) {
		HeModel model;
		Load3ds loader( argv[1], model );

		if( !model.empty()) {
			if( !model.writeIV( argv[2], "3ds2iv v0.1a" ))
				_gError("error: output couldnot be written\n"); 
		} else
			printf("empty model. no output written\n");
	} else
		printf("3ds2iv 0.1a by johannes zander - convert 3ds files to striped inventor ascii 2.1\nusage: 3ds2iv <input.3ds> <output.iv>\n");

	return( 0 );
}
